%%==============================================================================
%% Unit Tests for Code Navigation
%%==============================================================================
-module(els_definition_SUITE).

%% CT Callbacks
-export([
    all/0,
    init_per_suite/1,
    end_per_suite/1,
    init_per_testcase/2,
    end_per_testcase/2,
    suite/0
]).

%% Test cases
-export([
    application_local/1,
    application_remote/1,
    atom/1,
    behaviour/1,
    behaviour_callback_definition/1,
    definition_after_closing/1,
    duplicate_definition/1,
    export_entry/1,
    fun_local/1,
    fun_remote/1,
    import_entry/1,
    module_import_entry/1,
    include/1,
    include_lib/1,
    macro/1,
    macro_lowercase/1,
    macro_included/1,
    macro_with_args/1,
    macro_with_args_included/1,
    macro_with_implicit_args/1,
    multiple_atom_instances_same_mod/1,
    multiple_atom_instances_diff_mod/1,
    parse_transform/1,
    record_access/1,
    record_access_included/1,
    record_access_macro_name/1,
    record_expr/1,
    record_expr_included/1,
    record_expr_macro_name/1,
    record_field/1,
    record_field_included/1,
    record_type_macro_name/1,
    testcase/1,
    type_application_remote/1,
    type_application_undefined/1,
    type_application_user/1,
    type_export_entry/1,
    variable/1,
    variable_list_comp/1,
    opaque_application_remote/1,
    opaque_application_user/1,
    parse_incomplete/1
]).

%%==============================================================================
%% Includes
%%==============================================================================
-include_lib("common_test/include/ct.hrl").
-include_lib("stdlib/include/assert.hrl").

%%==============================================================================
%% Types
%%==============================================================================
-type config() :: [{atom(), any()}].

%%==============================================================================
%% CT Callbacks
%%==============================================================================
-spec all() -> [atom()].
all() ->
    els_test_utils:all(?MODULE).

-spec init_per_suite(config()) -> config().
init_per_suite(Config) ->
    els_test_utils:init_per_suite(Config).

-spec end_per_suite(config()) -> ok.
end_per_suite(Config) ->
    els_test_utils:end_per_suite(Config).

-spec init_per_testcase(atom(), config()) -> config().
init_per_testcase(TestCase, Config) ->
    els_test_utils:init_per_testcase(TestCase, Config).

-spec end_per_testcase(atom(), config()) -> ok.
end_per_testcase(TestCase, Config) ->
    els_test_utils:end_per_testcase(TestCase, Config).

-spec suite() -> [tuple()].
suite() ->
    [{timetrap, {seconds, 30}}].

%%==============================================================================
%% Testcases
%%==============================================================================
-spec application_local(config()) -> ok.
application_local(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 22, 5),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(Uri, DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {25, 1}, to => {25, 11}}),
        Range
    ),
    ok.

-spec application_remote(config()) -> ok.
application_remote(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 32, 13),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(?config(code_navigation_extra_uri, Config), DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {5, 1}, to => {5, 3}}),
        Range
    ),
    ok.

-spec atom(config()) -> ok.
atom(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def0 = els_client:definition(Uri, 84, 20),
    Def1 = els_client:definition(Uri, 85, 20),
    Def2 = els_client:definition(Uri, 86, 20),
    Def3 = els_client:definition(Uri, 85, 27),
    #{result := [#{range := Range0, uri := DefUri0}]} = Def0,
    #{result := [#{range := Range1, uri := DefUri1}, #{range := Range12, uri := DefUri12}]} =
        Def1,
    #{result := [#{range := Range2, uri := DefUri2}, #{range := Range22, uri := DefUri22}]} =
        Def2,
    #{result := [#{range := Range3, uri := DefUri3}]} = Def3,
    ?assertEqual(?config(code_navigation_types_uri, Config), DefUri0),
    ?assertEqual(
        els_protocol:range(#{from => {1, 9}, to => {1, 30}}),
        Range0
    ),
    ?assertEqual(?config(code_navigation_extra_uri, Config), DefUri12),
    ?assertEqual(
        els_protocol:range(#{from => {1, 9}, to => {1, 30}}),
        Range12
    ),
    ?assertEqual(Uri, DefUri1),
    ?assertEqual(
        els_protocol:range(#{from => {132, 1}, to => {132, 22}}),
        Range1
    ),
    ?assertEqual(?config(code_navigation_extra_uri, Config), DefUri22),
    ?assertEqual(
        els_protocol:range(#{from => {1, 9}, to => {1, 30}}),
        Range22
    ),
    ?assertEqual(Uri, DefUri2),
    ?assertEqual(
        els_protocol:range(#{from => {132, 1}, to => {132, 22}}),
        Range2
    ),
    ?assertEqual(?config('Code.Navigation.Elixirish_uri', Config), DefUri3),
    ?assertEqual(
        els_protocol:range(#{from => {1, 9}, to => {1, 36}}),
        Range3
    ),
    ok.

-spec behaviour(config()) -> ok.
behaviour(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 3, 16),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(?config(behaviour_a_uri, Config), DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {1, 9}, to => {1, 20}}),
        Range
    ),
    ok.

-spec behaviour_callback_definition(config()) -> ok.
behaviour_callback_definition(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 28, 5),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(?config(behaviour_a_uri, Config), DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {3, 1}, to => {3, 30}}),
        Range
    ),
    ok.

-spec testcase(config()) -> ok.
testcase(Config) ->
    Uri = ?config(sample_SUITE_uri, Config),
    Def = els_client:definition(Uri, 35, 6),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(Uri, DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {58, 1}, to => {58, 4}}),
        Range
    ),
    ok.

-spec multiple_atom_instances_same_mod(config()) -> ok.
multiple_atom_instances_same_mod(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Defs = els_client:definition(Uri, 130, 36),
    #{result := Results} = Defs,
    ?assertEqual(3, length(Results)),
    ExpectedRanges = [
        els_protocol:range(#{from => {1, 9}, to => {1, 24}}),
        els_protocol:range(#{from => {126, 1}, to => {126, 16}}),
        els_protocol:range(#{from => {128, 1}, to => {128, 16}})
    ],
    lists:foreach(
        fun(Def) ->
            #{range := Range, uri := DefUri} = Def,
            ?assertEqual(Uri, DefUri),
            ?assert(lists:member(Range, ExpectedRanges))
        end,
        Results
    ),
    ok.

-spec multiple_atom_instances_diff_mod(config()) -> ok.
multiple_atom_instances_diff_mod(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Defs = els_client:definition(Uri, 134, 35),
    #{result := Results} = Defs,
    ?assertEqual(2, length(Results)),
    RangeDef1 = els_protocol:range(#{from => {132, 1}, to => {132, 22}}),
    RangeDef2 = els_protocol:range(#{from => {1, 9}, to => {1, 30}}),
    Uri2 = ?config(code_navigation_extra_uri, Config),
    ?assertMatch(
        [
            #{
                range := RangeDef1,
                uri := Uri
            },
            #{
                range := RangeDef2,
                uri := Uri2
            }
        ],
        Results
    ),
    ok.

%% Issue #191: Definition not found after document is closed
-spec definition_after_closing(config()) -> ok.
definition_after_closing(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    ExtraUri = ?config(code_navigation_extra_uri, Config),
    Def = els_client:definition(Uri, 32, 13),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(ExtraUri, DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {5, 1}, to => {5, 3}}),
        Range
    ),

    %% Close file, get definition
    ok = els_client:did_close(ExtraUri),
    Def1 = els_client:definition(Uri, 32, 13),
    #{result := [#{range := Range, uri := DefUri}]} = Def1,
    ok.

-spec duplicate_definition(config()) -> ok.
duplicate_definition(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 57, 5),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(Uri, DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {60, 1}, to => {60, 11}}),
        Range
    ),
    ok.

-spec export_entry(config()) -> ok.
export_entry(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 8, 15),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(Uri, DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {28, 1}, to => {28, 11}}),
        Range
    ),
    ok.

-spec fun_local(config()) -> ok.
fun_local(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 51, 16),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(Uri, DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {25, 1}, to => {25, 11}}),
        Range
    ),
    ok.

-spec fun_remote(config()) -> ok.
fun_remote(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 52, 14),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(?config(code_navigation_extra_uri, Config), DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {5, 1}, to => {5, 3}}),
        Range
    ),
    ok.

-spec import_entry(config()) -> ok.
import_entry(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 10, 34),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(?config(code_navigation_extra_uri, Config), DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {5, 1}, to => {5, 3}}),
        Range
    ),
    ok.

-spec module_import_entry(config()) -> ok.
module_import_entry(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 90, 3),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(?config(code_navigation_extra_uri, Config), DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {5, 1}, to => {5, 3}}),
        Range
    ),
    ok.

-spec include(config()) -> ok.
include(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 12, 20),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(?config(code_navigation_h_uri, Config), DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {1, 1}, to => {1, 1}}),
        Range
    ),
    ok.

-spec include_lib(config()) -> ok.
include_lib(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 13, 22),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(?config(code_navigation_h_uri, Config), DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {1, 1}, to => {1, 1}}),
        Range
    ),
    ok.

-spec macro(config()) -> ok.
macro(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 26, 5),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(Uri, DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {18, 9}, to => {18, 16}}),
        Range
    ),
    ok.

-spec macro_lowercase(config()) -> ok.
macro_lowercase(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 48, 3),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(Uri, DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {45, 9}, to => {45, 16}}),
        Range
    ),
    ok.

-spec macro_included(config()) -> ok.
macro_included(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    UriHeader = ?config(code_navigation_h_uri, Config),
    #{result := [#{range := Range1, uri := DefUri1}]} =
        els_client:definition(Uri, 53, 19),
    ?assertEqual(UriHeader, DefUri1),
    ?assertEqual(
        els_protocol:range(#{from => {3, 9}, to => {3, 25}}),
        Range1
    ),
    #{result := [#{range := RangeQuoted, uri := DefUri2}]} =
        els_client:definition(Uri, 52, 75),
    ?assertEqual(UriHeader, DefUri2),
    ?assertEqual(
        els_protocol:range(#{from => {7, 9}, to => {7, 27}}),
        RangeQuoted
    ),
    ok.

-spec macro_with_args(config()) -> ok.
macro_with_args(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 40, 9),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(Uri, DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {19, 9}, to => {19, 16}}),
        Range
    ),
    ok.

-spec macro_with_args_included(config()) -> ok.
macro_with_args_included(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 43, 9),
    #{result := [#{uri := DefUri}]} = Def,
    ?assertEqual(
        <<"assert.hrl">>,
        filename:basename(els_uri:path(DefUri))
    ),
    %% Do not assert on line number to avoid binding to a specific OTP version
    ok.

-spec macro_with_implicit_args(config()) -> ok.
macro_with_implicit_args(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 124, 5),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(Uri, DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {118, 9}, to => {118, 16}}),
        Range
    ),
    ok.

-spec parse_transform(config()) -> ok.
parse_transform(Config) ->
    Uri = ?config(diagnostics_parse_transform_usage_uri, Config),
    Def = els_client:definition(Uri, 5, 45),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(?config(diagnostics_parse_transform_uri, Config), DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {1, 9}, to => {1, 36}}),
        Range
    ),
    ok.

-spec record_access(config()) -> ok.
record_access(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 34, 13),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(Uri, DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {16, 9}, to => {16, 17}}),
        Range
    ),
    ok.

-spec record_access_included(config()) -> ok.
record_access_included(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 52, 43),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(?config(code_navigation_h_uri, Config), DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {1, 9}, to => {1, 26}}),
        Range
    ),
    ok.

-spec record_access_macro_name(config()) -> ok.
record_access_macro_name(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 116, 33),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(Uri, DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {111, 9}, to => {111, 16}}),
        Range
    ),
    ok.

%% TODO: Additional constructors for POI
%% TODO: Navigation should return POI, not range
-spec record_expr(config()) -> ok.
record_expr(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 33, 11),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(Uri, DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {16, 9}, to => {16, 17}}),
        Range
    ),
    ok.

-spec record_expr_included(config()) -> ok.
record_expr_included(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 53, 30),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(?config(code_navigation_h_uri, Config), DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {1, 9}, to => {1, 26}}),
        Range
    ),
    ok.

-spec record_expr_macro_name(config()) -> ok.
record_expr_macro_name(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 115, 11),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(Uri, DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {111, 9}, to => {111, 16}}),
        Range
    ),
    ok.

-spec record_field(config()) -> ok.
record_field(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 33, 20),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(Uri, DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {16, 20}, to => {16, 27}}),
        Range
    ),
    ok.

-spec record_field_included(config()) -> ok.
record_field_included(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 53, 45),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(?config(code_navigation_h_uri, Config), DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {1, 29}, to => {1, 45}}),
        Range
    ),
    ok.

-spec record_type_macro_name(config()) -> ok.
record_type_macro_name(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 113, 28),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(Uri, DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {111, 9}, to => {111, 16}}),
        Range
    ),
    ok.

-spec type_application_remote(config()) -> ok.
type_application_remote(Config) ->
    ExtraUri = ?config(code_navigation_extra_uri, Config),
    TypesUri = ?config(code_navigation_types_uri, Config),
    Def = els_client:definition(ExtraUri, 11, 38),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(TypesUri, DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {3, 1}, to => {3, 26}}),
        Range
    ),
    ok.

-spec type_application_undefined(config()) -> ok.
type_application_undefined(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 55, 42),
    #{result := Result} = Def,
    Expected = [
        #{
            range => #{
                'end' => #{character => 49, line => 54},
                start => #{character => 33, line => 54}
            },
            uri => Uri
        }
    ],
    ?assertEqual(Expected, Result),
    ok.

-spec type_application_user(config()) -> ok.
type_application_user(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 55, 25),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(Uri, DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {37, 1}, to => {37, 25}}),
        Range
    ),
    ok.

-spec type_export_entry(config()) -> ok.
type_export_entry(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    Def = els_client:definition(Uri, 9, 17),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(Uri, DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {37, 1}, to => {37, 25}}),
        Range
    ),
    ok.

-spec variable(config()) -> ok.
variable(Config) ->
    Uri = ?config(code_navigation_uri, Config),
    ?assertEqual(
        {#{from => {103, 12}, to => {103, 15}}, Uri},
        definition(Uri, 104, 9)
    ),
    ?assertEqual(
        {#{from => {104, 3}, to => {104, 6}}, Uri},
        definition(Uri, 105, 10)
    ),
    ?assertEqual(
        {#{from => {106, 12}, to => {106, 15}}, Uri},
        definition(Uri, 107, 10)
    ),
    ?assertEqual(
        {#{from => {106, 12}, to => {106, 15}}, Uri},
        definition(Uri, 108, 10)
    ),
    %% Inside macro
    ?assertEqual(
        {#{from => {19, 17}, to => {19, 18}}, Uri},
        definition(Uri, 19, 36)
    ),
    ok.

-spec variable_list_comp(config()) -> ok.
variable_list_comp(Config) ->
    Uri = ?config(variable_list_comp_uri, Config),

    %% one: Should skip LC
    ?assertEqual(
        {#{from => {4, 5}, to => {4, 8}}, Uri},
        definition(Uri, 6, 5)
    ),
    %% one: Should go to LC generator pattern
    ?assertEqual(
        {#{from => {5, 14}, to => {5, 17}}, Uri},
        definition(Uri, 5, 7)
    ),
    %% two: Should go to first LC generator pattern
    ?assertEqual(
        {#{from => {9, 14}, to => {9, 17}}, Uri},
        definition(Uri, 9, 7)
    ),
    %% two: Should go to second LC generator pattern
    ?assertEqual(
        {#{from => {10, 14}, to => {10, 17}}, Uri},
        definition(Uri, 10, 7)
    ),
    %% three: Should go to variable definition
    ?assertEqual(
        {#{from => {13, 5}, to => {13, 8}}, Uri},
        definition(Uri, 14, 7)
    ),
    %% three: Should go to variable definition (same)
    ?assertEqual(
        {#{from => {13, 5}, to => {13, 8}}, Uri},
        definition(Uri, 15, 5)
    ),
    %% four: Should go to first LC generator pattern
    ?assertEqual(
        {#{from => {18, 22}, to => {18, 25}}, Uri},
        definition(Uri, 18, 8)
    ),
    %% four: Should go to second LC generator pattern
    ?assertEqual(
        {#{from => {19, 22}, to => {19, 26}}, Uri},
        definition(Uri, 18, 13)
    ),
    %% four: Should go to third LC generator pattern
    ?assertEqual(
        {#{from => {19, 39}, to => {19, 42}}, Uri},
        definition(Uri, 19, 32)
    ),
    ok.

-spec opaque_application_remote(config()) -> ok.
opaque_application_remote(Config) ->
    ExtraUri = ?config(code_navigation_extra_uri, Config),
    TypesUri = ?config(code_navigation_types_uri, Config),
    Def = els_client:definition(ExtraUri, 16, 61),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(TypesUri, DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {7, 1}, to => {7, 35}}),
        Range
    ),
    ok.

-spec opaque_application_user(config()) -> ok.
opaque_application_user(Config) ->
    ExtraUri = ?config(code_navigation_extra_uri, Config),
    Def = els_client:definition(ExtraUri, 16, 24),
    #{result := [#{range := Range, uri := DefUri}]} = Def,
    ?assertEqual(ExtraUri, DefUri),
    ?assertEqual(
        els_protocol:range(#{from => {20, 1}, to => {20, 34}}),
        Range
    ),
    ok.

-spec parse_incomplete(config()) -> ok.
parse_incomplete(Config) ->
    Uri = ?config(code_navigation_broken_uri, Config),
    Range = els_protocol:range(#{from => {3, 1}, to => {3, 11}}),
    ?assertMatch(
        #{result := [#{range := Range, uri := Uri}]},
        els_client:definition(Uri, 7, 3)
    ),
    ?assertMatch(
        #{result := [#{range := Range, uri := Uri}]},
        els_client:definition(Uri, 8, 3)
    ),
    ?assertMatch(
        #{result := [#{range := Range, uri := Uri}]},
        els_client:definition(Uri, 9, 8)
    ),
    ?assertMatch(
        #{result := [#{range := Range, uri := Uri}]},
        els_client:definition(Uri, 11, 7)
    ),
    ?assertMatch(
        #{result := [#{range := Range, uri := Uri}]},
        els_client:definition(Uri, 12, 12)
    ),
    ?assertMatch(
        #{result := [#{range := Range, uri := Uri}]},
        els_client:definition(Uri, 17, 3)
    ),
    ?assertMatch(
        #{result := [#{range := Range, uri := Uri}]},
        els_client:definition(Uri, 19, 3)
    ),
    ok.

definition(Uri, Line, Char) ->
    case els_client:definition(Uri, Line, Char) of
        #{result := [#{range := Range, uri := DefUri}]} ->
            {els_range:to_poi_range(Range), DefUri};
        Res ->
            {error, Res}
    end.
